"""Site Model List Item."""
from __future__ import annotations

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.containers import Vertical
from textual.widgets import ListItem
from textual.widgets import Static

from parllama.models.ollama_data import SiteModel


class SiteModelListItem(ListItem):
    """Site Model List Item."""

    DEFAULT_CSS = """
    SiteModelListItem {
        padding: 1 2 0 2;
        width: 1fr;
        min-height: 7;
        max-height: 8;
        background: $background;
        border: solid $accent;
        border-title-color: $primary;

        Static {
            width: 1fr;
            max-height: 3;
        }
        Static.tags {
            link-background: transparent;
            link-color: #2f82ff;
            link-color-hover: #4b93fd;
            link-background-hover: transparent;
        }
        #updated {
            align: right top;
            dock: bottom;
            height: 1;
            width: auto;
            color: $text-muted;
        }
    }
    """
    model: SiteModel

    def __init__(self, model: SiteModel) -> None:
        """Initialize the widget."""
        super().__init__()
        self.model = model
        self.can_focus = True
        self.border_title = self.model.name

    def watch_has_focus(self, value: bool) -> None:
        """Watch the has_focus property."""
        super().watch_has_focus(value)
        if value and hasattr(self.parent, "selected"):
            self.parent.selected = self if value else None  # type: ignore

    def compose(self) -> ComposeResult:
        """Compose the item."""
        with Vertical():
            yield Static(self.model.description, id="description")
            with Horizontal():
                with Horizontal():
                    yield Static(
                        "Tags: "
                        + " ".join(
                            [
                                f"[@click=screen.site_tag_clicked('{self.model.name}:{t.lower()}')]{t}[/]"
                                for t in self.model.tags
                            ]
                        ),
                        classes="tags",
                    )
                yield Static(f"Updated: {self.model.updated}", id="updated")



"""Session list side panel"""
from __future__ import annotations

from typing import cast

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import VerticalScroll
from textual.widgets import ListView

from parllama.chat_manager import chat_manager
from parllama.messages.main import DeleteSession
from parllama.messages.main import SessionListChanged
from parllama.widgets.session_list_item import SessionListItem


class SessionList(VerticalScroll, can_focus=False, can_focus_children=True):
    """Session list side panel"""

    DEFAULT_CSS = """
    """
    BINDINGS = [
        Binding(
            key="enter",
            action="load_item",
            description="Load",
            show=True,
            priority=True,
        ),
        Binding(
            key="delete",
            action="delete_item",
            description="Delete",
            show=True,
        ),
    ]
    list_view: ListView

    def __init__(self, **kwargs) -> None:
        """Initialise the view."""
        super().__init__(**kwargs)
        self.list_view = ListView(initial_index=None)

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""
        with self.list_view:
            for s in chat_manager.sessions:
                if not s.is_valid():
                    continue
                yield SessionListItem(s)

    def on_show(self) -> None:
        """Initialise the view."""
        if not chat_manager.current_session:
            self.notify("no session")
            return
        for item in self.list_view.query(SessionListItem):
            if item.session.session_id == chat_manager.current_session.session_id:
                self.list_view.index = self.list_view.children.index(item)
                break

    def action_delete_item(self) -> None:
        """Handle delete item action."""
        selected_item: SessionListItem = cast(
            SessionListItem, self.list_view.highlighted_child
        )
        if not selected_item:
            return
        self.app.post_message(
            DeleteSession(session_id=selected_item.session.session_id)
        )
        self.display = False

    @on(SessionListChanged)
    async def on_session_list_changed(self, event: SessionListChanged) -> None:
        """Handle session list changed event."""
        event.stop()
        await self.recompose()

    def action_load_item(self) -> None:
        """Handle list view selected event."""
        selected_item: SessionListItem = cast(
            SessionListItem, self.list_view.highlighted_child
        )
        if not selected_item:
            return
        chat_manager.set_current_session(selected_item.session.session_id)
        self.display = False



"""Filter Input with custom bindings"""
from __future__ import annotations

from typing import ClassVar

from textual.binding import Binding
from textual.binding import BindingType
from textual.widgets import Input


class FilterInput(Input, inherit_bindings=False):
    """Filter Input with custom bindings"""

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("left", "cursor_left", "cursor left", show=False),
        Binding("ctrl+left", "cursor_left_word", "cursor left word", show=False),
        Binding("right", "cursor_right", "cursor right", show=False),
        Binding("ctrl+right", "cursor_right_word", "cursor right word", show=False),
        Binding("backspace", "delete_left", "delete left", show=False),
        Binding("home,ctrl+a", "home", "home", show=False),
        Binding("end,ctrl+e", "end", "end", show=False),
        Binding("delete,ctrl+d", "delete_right", "delete right", show=False),
        Binding("enter", "submit", "submit", show=False),
    ]



"""Widget for chatting with LLM."""
from __future__ import annotations

from typing import cast

from ollama import Options
from textual import on
from textual import work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container
from textual.containers import Horizontal
from textual.containers import Vertical
from textual.containers import VerticalScroll
from textual.events import Focus
from textual.events import Show
from textual.message import Message
from textual.reactive import Reactive
from textual.suggester import SuggestFromList
from textual.widgets import Button
from textual.widgets import Input
from textual.widgets import Label
from textual.widgets import Select
from textual.widgets import TabbedContent

from parllama.chat_manager import chat_manager
from parllama.data_manager import dm
from parllama.dialogs.information import InformationDialog
from parllama.messages.main import ChatMessage
from parllama.messages.main import ChatMessageSent
from parllama.messages.main import DeleteSession
from parllama.messages.main import LocalModelListLoaded
from parllama.messages.main import RegisterForUpdates
from parllama.messages.main import SessionSelected
from parllama.models.chat import ChatSession
from parllama.models.chat import OllamaMessage
from parllama.models.settings_data import settings
from parllama.screens.save_session import SaveSession
from parllama.widgets.chat_message import ChatMessageWidget
from parllama.widgets.input_tab_complete import InputTabComplete
from parllama.widgets.session_list import SessionList

valid_commands: list[str] = [
    "/new",
    "/delete",
    "/export",
    "/model",
    "/temp",
    "/session",
    "/help",
    "/?",
]


class ChatView(Container, can_focus=False, can_focus_children=True):
    """Widget for viewing application logs."""

    DEFAULT_CSS = """
    ChatView {
      layers: left;
      #tool_bar {
        height: 3;
        background: $surface-darken-1;
        #model_name {
          max-width: 40;
        }
        #temperature_input {
          width: 11;
        }
        #session_name_input {
          min-width: 15;
          max-width: 40;
          width: auto;
        }
        #new_button {
          margin-left: 2;
          min-width: 9;
          background: $warning-darken-2;
          border-top: tall $warning-lighten-1;
        }
        Label {
          margin: 1;
          background: transparent;
        }
      }
      #send_bar {
        height: 3;
        background: $surface-darken-1;
        #user_input {
          width: 1fr;
        }
        #send_button {
          width: 6;
        }
      }
      #messages {
        background: $primary-background;
        ChatMessageWidget{
            padding: 1;
            border: none;
            border-left: blank;
            &:focus {
                border-left: thick $primary;
            }
        }
        MarkdownH2 {
          margin: 0;
          padding: 0;
        }
      }
    }
    """

    BINDINGS = [
        Binding(
            key="ctrl+s",
            action="toggle_session_list",
            description="Sessions",
            show=True,
        ),
        Binding(
            key="ctrl+n",
            action="new_session",
            description="New",
            show=True,
        ),
    ]
    session: ChatSession
    busy: Reactive[bool] = Reactive(False)

    def __init__(self, **kwargs) -> None:
        """Initialise the view."""
        super().__init__(**kwargs)

        self.model_select: Select[str] = Select(
            id="model_name", options=[], prompt="Select Model"
        )
        self.temperature_input: Input = Input(
            id="temperature_input",
            value=(
                f"{settings.last_chat_temperature:.2f}"
                if settings.last_chat_temperature
                else ""
            ),
            max_length=4,
            restrict=r"^\d?\.?\d?$",
        )
        self.session_name_input: Input = Input(
            id="session_name_input", value=settings.last_chat_session_name or "New Chat"
        )
        self.user_input: InputTabComplete = InputTabComplete(
            id="user_input",
            placeholder="Type a message...",
            suggester=SuggestFromList(
                valid_commands,
                case_sensitive=False,
            ),
            submit_on_tab=False,
            submit_on_complete=False,
        )

        self.send_button: Button = Button("Send", id="send_button", disabled=True)
        self.vs: VerticalScroll = VerticalScroll(id="messages")
        self.vs.can_focus = False
        self.busy = False
        self.session_list = SessionList()
        self.session_list.display = False

        self.session = chat_manager.get_or_create_session_name(
            session_name=self.session_name_input.value or "New Chat",
            model_name=str(self.model_select.value),
            options=self.get_session_options(),
        )

    def _watch_busy(self, value: bool) -> None:
        """Update controls when busy changes"""
        self.update_control_states()
        if value:
            self.screen.sub_title = (  # pylint: disable=attribute-defined-outside-init
                "Chat - Thinking..."
            )
        else:
            self.screen.sub_title = (  # pylint: disable=attribute-defined-outside-init
                "Chat"
            )

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""
        yield self.session_list
        with Vertical(id="main"):
            with Horizontal(id="tool_bar"):
                yield self.model_select
                yield Label("Temp")
                yield self.temperature_input
                yield Label("Session")
                yield self.session_name_input
                yield Button("New", id="new_button", variant="warning")
            with self.vs:
                yield from [
                    ChatMessageWidget.mk_msg_widget(msg=m)
                    for m in self.session.messages
                ]
            with Horizontal(id="send_bar"):
                yield self.user_input
                yield self.send_button

    async def on_mount(self) -> None:
        """Set up the dialog once the DOM is ready."""
        self.app.post_message(RegisterForUpdates(widget=self))

    def _on_show(self, event: Show) -> None:
        """Handle show event"""
        self._watch_busy(self.busy)
        with self.screen.prevent(TabbedContent.TabActivated):
            if self.model_select.value == Select.BLANK:
                self.model_select.focus()
            else:
                self.user_input.focus()

    @on(Input.Changed, "#user_input")
    def on_user_input_changed(self) -> None:
        """Handle max lines input change"""
        self.update_control_states()

    @on(Input.Changed, "#temperature_input")
    def on_temperature_input_changed(self) -> None:
        """Handle temperature input change"""
        try:
            if self.temperature_input.value:
                settings.last_chat_temperature = float(self.temperature_input.value)
            else:
                settings.last_chat_temperature = None
        except ValueError:
            return
        self.session.set_temperature(settings.last_chat_temperature)
        settings.save_settings_to_file()
        chat_manager.notify_changed()

    @on(Input.Changed, "#session_name_input")
    def on_session_name_input_changed(self) -> None:
        """Handle session name input change"""
        session_name: str = self.session_name_input.value.strip()
        if not session_name:
            return
        settings.last_chat_session_name = session_name
        settings.save_settings_to_file()

        self.session.set_name(settings.last_chat_session_name)
        chat_manager.notify_changed()

    def update_control_states(self):
        """Update disabled state of controls based on model and user input values"""
        self.send_button.disabled = (
            self.busy
            or self.model_select.value == Select.BLANK
            or len(self.user_input.value.strip()) == 0
        )

    @on(Select.Changed)
    def on_model_select_changed(self) -> None:
        """Model select changed, update control states and save model name"""
        self.update_control_states()
        if self.model_select.value not in (Select.BLANK, settings.last_chat_model):
            settings.last_chat_model = str(self.model_select.value)
            settings.save_settings_to_file()
            self.session.set_llm_model(self.model_select.value)  # type: ignore

    def set_model_name(self, model_name: str) -> None:
        """ "Set model names"""
        for _, v in dm.get_model_select_options():
            if v == model_name:
                self.model_select.value = model_name
                return
        self.model_select.value = Select.BLANK

    @on(LocalModelListLoaded)
    def on_local_model_list_loaded(self, evt: LocalModelListLoaded) -> None:
        """Model list changed"""
        evt.stop()

        self.user_input.suggester = SuggestFromList(
            valid_commands + [f"/model {m.model.name}" for m in dm.models],
            case_sensitive=False,
        )

        if self.model_select.value != Select.BLANK:
            old_v = self.model_select.value
        elif settings.last_chat_model:
            old_v = settings.last_chat_model
        else:
            old_v = Select.BLANK
        opts = dm.get_model_select_options()
        self.model_select.set_options(opts)
        for _, v in opts:
            if v == old_v:
                self.model_select.value = old_v

        self.update_control_states()

        if self.model_select.value != Select.BLANK:
            self.session.set_llm_model(self.model_select.value)  # type: ignore
            #  TODO fix this smell
            if (
                self.parent
                and self.parent.parent
                and self.parent.parent.parent
                and cast(TabbedContent, self.parent.parent.parent).active == "Chat"
            ):
                self.user_input.focus()

    @on(Button.Pressed, "#new_button")
    def on_new_button_pressed(self, event: Button.Pressed) -> None:
        """New button pressed"""
        event.stop()
        self.action_new_session()

    def get_session_options(self) -> Options | None:
        """Get session options"""
        if not self.temperature_input.value:
            return None
        return {"temperature": (float(self.temperature_input.value))}

    def action_new_session(self) -> None:
        """Start new session"""
        with self.prevent(Input.Changed):
            self.session = chat_manager.new_session(
                session_name="New Chat",
                model_name=str(self.model_select.value),
                options=self.get_session_options(),
            )
            self.session_name_input.value = self.session.session_name

        self.vs.remove_children("*")
        self.update_control_states()
        self.user_input.focus()

    @on(Button.Pressed, "#send_button")
    @on(Input.Submitted, "#user_input")
    async def action_send_message(self, event: Message) -> None:
        """Send the message."""
        event.stop()
        if self.model_select.value == Select.BLANK:
            self.model_select.focus()
            return

        self.user_input.focus()

        if self.send_button.disabled:
            return

        user_msg: str = self.user_input.value.strip()
        if not user_msg:
            return

        if self.busy:
            self.notify("LLM Busy...", severity="error")
            return

        self.update_control_states()
        self.user_input.value = ""
        if user_msg.startswith("/"):
            return self.handle_command(user_msg[1:].lower().strip())
        self.busy = True
        self.do_send_message(user_msg)

    # pylint: disable=too-many-branches
    def handle_command(self, cmd: str) -> None:
        """Handle a command"""
        if cmd in ("?", "help"):
            self.app.push_screen(
                InformationDialog(
                    title="Chat Commands",
                    message="""
Chat Commands:
/new - New a chat session
/delete - Delete current chat session and create new one
/model [model_name] - Select a model
/temp [temperature] - Set the temperature
/session [session_name] - Set the session name
/export - Export the conversation to a Markdown file
                    """,
                )
            )
            return

        if self.session is None:
            return
        if cmd == "new":
            self.action_new_session()
        if cmd == "delete":
            self.app.post_message(DeleteSession(session_id=self.session.session_id))
        elif cmd == "model":
            self.set_timer(0.1, self.model_select.focus)
            return
        elif cmd.startswith("model "):
            (_, v) = cmd.split(" ", 1)
            if v not in [m.model.name for m in dm.models]:
                self.notify(f"Model {v} not found", severity="error")
                return
            self.model_select.value = v
            self.set_timer(0.1, self.user_input.focus)
        elif cmd == "temp":
            self.set_timer(0.1, self.temperature_input.focus)
        elif cmd.startswith("temp "):
            (_, v) = cmd.split(" ", 1)
            self.temperature_input.value = v
            self.set_timer(0.1, self.user_input.focus)
        elif cmd == "session":
            self.set_timer(0.1, self.session_name_input.focus)
        elif cmd.startswith("session "):
            (_, v) = cmd.split(" ", 1)
            self.session_name_input.value = v
            self.set_timer(0.1, self.user_input.focus)
        elif cmd.startswith("export"):
            self._save_conversation_text()
            # filename: str = "chat_test.md"
            # if self.session.save(filename):
            #     self.notify(f"Saved: {filename}")
            # else:
            #     self.notify(f"Failed to save: {filename}", severity="error")
        else:
            self.notify(f"Unknown command: {cmd}", severity="error")

    @work(thread=True)
    async def do_send_message(self, msg: str) -> None:
        """Send the message."""

        if self.session is None:
            return
        await self.session.send_chat(msg, self)
        self.post_message(ChatMessageSent())

    @on(ChatMessageSent)
    def on_chat_message_sent(self, event: ChatMessageSent) -> None:
        """Handle a chat message sent"""
        event.stop()
        self.busy = False

    @on(ChatMessage)
    async def on_chat_message(self, event: ChatMessage) -> None:
        """Handle a chat message"""
        event.stop()

        ses = chat_manager.get_session(event.session_id)
        if not ses:
            self.notify("Chat session not found", severity="error")
            return
        msg: OllamaMessage | None = ses.get_message(event.message_id)
        if not msg:
            self.notify("Chat message not found", severity="error")
            return

        msg_widget: ChatMessageWidget | None = None
        for w in self.query(ChatMessageWidget):
            if w.msg.message_id == msg.message_id:
                await w.update("")
                msg_widget = w
                break
        if not msg_widget:
            msg_widget = ChatMessageWidget.mk_msg_widget(msg=msg)
            await self.vs.mount(msg_widget)
        msg_widget.loading = len(msg_widget.msg.content) == 0
        if self.user_input.has_focus:
            self.set_timer(0.05, self.scroll_to_bottom)
        chat_manager.notify_changed()

    def scroll_to_bottom(self) -> None:
        """Scroll to the bottom of the chat window."""
        self.vs.scroll_to(y=self.vs.scroll_y + self.vs.size.height)

    @work
    async def _save_conversation_text(self) -> None:
        """Save the conversation as a Markdown document."""
        # Prompt the user with a save dialog, to get the name of a file to
        # save to.
        if (target := await SaveSession.get_filename(self.app)) is None:
            return

        # If the user didn't give an extension, add a default.
        if not target.suffix:
            target = target.with_suffix(".md")

        # Save the Markdown to the target file.
        target.write_text(str(self.session), encoding="utf-8")

        # Let the user know the save happened.
        self.notify(str(target), title="Saved")

    def action_toggle_session_list(self) -> None:
        """Toggle the session list."""
        self.session_list.display = not self.session_list.display
        if self.session_list.display:
            self.set_timer(0.1, self.session_list.list_view.focus)

    def load_session(self, session_id: str) -> None:
        """Load a session"""
        session = chat_manager.get_session(session_id)
        if not session:
            self.notify("Chat session not found", severity="error")
            return
        self.session = session
        self.vs.remove_children("*")
        self.vs.mount(
            *[ChatMessageWidget.mk_msg_widget(msg=m) for m in self.session.messages]
        )
        with self.prevent(Focus, Input.Changed, Select.Changed):
            self.set_model_name(self.session.llm_model_name)
            self.temperature_input.value = str(
                self.session.options.get("temperature", "")
            )
            self.session_name_input.value = self.session.session_name
        self.update_control_states()
        self.set_timer(0.05, self.scroll_to_bottom)
        self.user_input.focus()

    @on(SessionSelected)
    def on_session_selected(self, event: SessionSelected) -> None:
        """Session selected event"""
        event.stop()
        self.load_session(event.session_id)

    @on(DeleteSession)
    def on_delete_session(self, event: DeleteSession) -> None:
        """Delete session event"""
        event.stop()
        chat_manager.delete_session(event.session_id)
        self.notify("Chat session deleted")
        if self.session.session_id == event.session_id:
            self.action_new_session()



"""Data manager for Par Llama."""
from __future__ import annotations

import os.path
import re
import shutil
from collections.abc import Generator
from collections.abc import Iterator
from collections.abc import Mapping
from datetime import datetime
from typing import Any
from typing import Literal

import docker.errors  # type: ignore
import docker.types  # type: ignore
import requests
import simplejson as json
from bs4 import BeautifulSoup
from docker.models.containers import Container  # type: ignore

from parllama.docker_utils import start_docker_container
from parllama.models.ollama_data import FullModel
from parllama.models.ollama_data import ModelListPayload
from parllama.models.ollama_data import ModelShowPayload
from parllama.models.ollama_data import SiteModel
from parllama.models.ollama_data import SiteModelData
from parllama.models.settings_data import settings
from parllama.utils import output_to_dicts
from parllama.utils import run_cmd
from parllama.widgets.local_model_list_item import LocalModelListItem
from parllama.widgets.site_model_list_item import SiteModelListItem


class DataManager:
    """Data manager for Par Llama."""

    ollama_site_categories: list[str] = ["popular", "featured", "newest"]
    models: list[LocalModelListItem]
    site_models: list[SiteModelListItem]
    ollama_bin: str

    def __init__(self):
        """ "Initialize the data manager."""
        self.models = []
        self.site_models = []
        # get location of ollama binary in path
        ollama_bin = shutil.which("ollama") or shutil.which("ollama.exe")
        if not ollama_bin:
            raise FileNotFoundError("Could not find ollama binary in path")

        self.ollama_bin = str(ollama_bin)

    @staticmethod
    def _get_all_model_data() -> list[LocalModelListItem]:
        """Get all model data."""
        all_models: list[LocalModelListItem] = []
        res = ModelListPayload(**settings.ollama_client.list())
        pattern = r"^(# Modelfile .*)\n(# To build.*)\n# (FROM .*\n)\n(FROM .*)\n(.*)$"
        replacement = r"\3\5"
        for model in res.models:
            res2 = ModelShowPayload(**settings.ollama_client.show(model.name))

            res2.modelfile = re.sub(
                pattern, replacement, res2.modelfile, flags=re.MULTILINE | re.IGNORECASE
            )
            res3 = FullModel(**model.model_dump(), **res2.model_dump())
            all_models.append(LocalModelListItem(res3))
        return all_models

    def refresh_models(self) -> list[LocalModelListItem]:
        """Refresh all local model data."""
        self.models = self._get_all_model_data()
        return self.models

    def get_model_select_options(self) -> list[tuple[str, str]]:
        """Get select options."""
        return [(model.model.name, model.model.name) for model in self.models]

    def model_ps(self) -> list[dict[str, Any]]:
        """Get model ps."""
        ret = run_cmd([self.ollama_bin, "ps"])

        if not ret:
            return []
        return output_to_dicts(ret)

    @staticmethod
    def pull_model(model_name: str) -> Iterator[dict[str, Any]]:
        """Pull a model."""
        return settings.ollama_client.pull(model_name, stream=True)  # type: ignore

    @staticmethod
    def push_model(model_name: str) -> Iterator[dict[str, Any]]:
        """Push a model."""
        return settings.ollama_client.push(model_name, stream=True)  # type: ignore

    def delete_model(self, model_name: str) -> bool:
        """Delete a model."""
        ret = settings.ollama_client.delete(model_name).get("status", False)
        # ret = True
        if not ret:
            return False

        for model in self.models:
            if model.model.name == model_name:
                # model.remove()
                self.models.remove(model)
                return True

        return False

    def list_cache_files(self) -> list[str]:
        """List cache files."""
        if not os.path.exists(settings.cache_dir):
            return []

        return [
            f.split("-")[1].split(".")[0]
            for f in os.listdir(settings.cache_dir)
            if os.path.isfile(os.path.join(settings.cache_dir, f))
            and f.lower().endswith(".json")
        ]

    # pylint: disable=too-many-branches
    def refresh_site_models(
        self,
        namespace: str,
        category: Literal["popular", "featured", "newest"] | None = None,
        force: bool = True,
    ) -> list[SiteModelListItem]:
        """Get list of all available models from Ollama.com."""

        settings.ensure_cache_folder()

        if not namespace:
            namespace = "models"
        namespace = os.path.basename(namespace)

        file_name = os.path.join(settings.cache_dir, f"site_models-{namespace}.json")
        if not force and os.path.exists(file_name):
            try:
                with open(file_name, encoding="utf-8") as f:
                    ret: SiteModelData = SiteModelData(**json.load(f))
                    if ret.last_update and ret.last_update.timestamp() > (
                        ret.last_update.timestamp() - 60 * 60 * 24
                    ):
                        self.site_models = [SiteModelListItem(m) for m in ret.models]
                        return self.site_models
            except Exception as e:  # pylint: disable=broad-exception-caught
                print(f"Error: {e}")

        url_base: str = f"https://ollama.com/{namespace}"
        models: list[SiteModel] = []

        for cat in self.ollama_site_categories:
            if category and category != cat:
                continue
            url = url_base
            if namespace == "models":
                url += f"?sort={cat}"
            response = requests.get(
                url, headers={"User-Agent": "Mozilla/5.0"}, timeout=10
            )
            soup = BeautifulSoup(response.content.decode("utf-8"), "html.parser")
            for card in soup.find_all("li", class_="items-baseline"):
                meta_data = {
                    "name": card.find("h2").text.strip(),
                    "description": card.find("p").text.strip(),
                    "model_url": card.find("a")["href"],
                    "url": f'{url}{card.find("a")["href"]}',
                    "num_pulls": "",
                    "num_tags": "",
                    "updated": "",
                }

                pres = card.find_all(
                    "span", class_=["flex", "items-center"], recursive=True
                )
                pres = [p.text.strip() for p in pres]
                pres = [p for p in pres if p]

                for p in pres:
                    if "Pulls" in p:
                        meta_data["num_pulls"] = p.split("\n")[0].strip()
                    if "Tags" in p:
                        meta_data["num_tags"] = p.split("\xa0")[0].strip()
                    if "Updated" in p:
                        meta_data["updated"] = p.split("\xa0")[-1].strip()
                # tags = []
                tags = card.find_all("span", class_=["text-blue-600"], recursive=True)
                tags = [p.text.strip() for p in tags]
                tags = [p for p in tags if p]
                meta_data["tags"] = tags

                model = SiteModel(**meta_data)
                # print(model.model_dump_json(indent=4))
                found = False
                for m in models:
                    if m.name == model.name:
                        found = True
                        break
                if not found:
                    models.append(model)

            if namespace != "models":
                break

        if len(models) > 0 and not settings.no_save:
            with open(file_name, "w", encoding="utf-8") as f:
                f.write(
                    SiteModelData(
                        models=models, last_update=datetime.now()
                    ).model_dump_json(indent=4)
                )
        self.site_models = [SiteModelListItem(m) for m in models]
        return self.site_models

    @staticmethod
    def create_model(
        model_name: str,
        model_code: str,
        quantize_level: str | None = None,
    ) -> Iterator[dict[str, Any]]:
        """Create a new model."""
        return settings.ollama_client.create(
            model=model_name,
            modelfile=model_code,
            quantize=quantize_level,
            stream=True,
        )  # type: ignore

    @staticmethod
    def copy_model(src_name: str, dst_name: str) -> Mapping[str, Any]:
        """Copy local model to new name"""
        return settings.ollama_client.copy(source=src_name, destination=dst_name)

    @staticmethod
    def quantize_model(
        model_name: str, quantize_level: str = "q4_0"
    ) -> str | Container | Generator[bytes, None, None]:
        """
        Quantize a model

        Args:
            model_name (str): The name of the model to quantize
            quantize_level (str): The quantization level to use. Defaults to "q4_0"

        Returns:
            bool: True if successful, False otherwise
        """
        model_name = os.path.basename(model_name)
        model_folder = os.path.join(settings.data_dir, "quantize_workspace", model_name)
        if not os.path.exists(model_folder):
            raise FileNotFoundError(f"Model folder does not exist: {model_folder}")
        quantized_model_file = os.path.join(model_folder, f"{quantize_level}.bin")
        if os.path.exists(quantized_model_file):
            os.unlink(quantized_model_file)

        # docker run --rm -v .:/model ollama/quantize -q q4_0 /model
        ret = start_docker_container(
            image="ollama/quantize",
            container_name="parllamaQtz",
            command=f"-q {quantize_level} /model",
            network_name="parllama-network",
            mounts=[
                docker.types.Mount(
                    target="/model",
                    source=model_folder,
                    type="bind",
                )
            ],
            background=False,
        )
        if isinstance(ret, Exception):
            raise ret

        if not os.path.exists(quantized_model_file):
            raise FileNotFoundError(
                f"Quantized model does not exist: {quantized_model_file}"
            )
        if isinstance(ret, str):
            return ret

        return ret.logs(stream=True)


dm: DataManager = DataManager()



"""Toggleable hidden text input field."""
from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import Button
from textual.widgets import Input

from parllama.icons import EYE_EMOJI
from parllama.utils import to_class_case


class HiddenInputField(Widget):
    """Toggleable hidden text input field."""

    DEFAULT_CSS = """
    HiddenInputField {
        height: 3;
        width: 1fr;
        Input {
            width: 1fr;
        }
    }
    """
    base_name: str
    input: Input

    def __init__(self, base_name: str, **kwargs) -> None:
        """Initialize the Hidden Field."""
        base_name = to_class_case(base_name.replace("/", ""))
        super().__init__(name=f"{base_name}HiddenField")

        self.base_name = base_name
        self.input = Input(id=f"{self.base_name}Input", password=True, **kwargs)

    def compose(self) -> ComposeResult:
        """Compose the list item."""
        with Horizontal(classes="width-fr-1"):
            yield self.input
            btn = Button(EYE_EMOJI, id="show", classes="field-button")
            btn.tooltip = "Show/Hide"
            yield btn

    @on(Button.Pressed, "#show")
    def show(self, event: Button.Pressed) -> None:
        """Toggle the hidden input field."""
        event.stop()
        ctrl = self.query_one(Input)
        ctrl.password = not ctrl.password

    @property
    def value(self) -> str:
        """Get the value."""
        return self.input.value

    @value.setter
    def value(self, value: str) -> None:
        """Set the value."""
        self.input.value = value



"""Widget for viewing application logs."""
from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.containers import Container
from textual.containers import Horizontal
from textual.containers import Vertical
from textual.events import Show
from textual.widgets import Button
from textual.widgets import Checkbox
from textual.widgets import Input
from textual.widgets import Label
from textual.widgets import RichLog

from parllama.models.settings_data import settings


class LogView(Container):
    """Widget for viewing application logs."""

    DEFAULT_CSS = """
    LogView {
      #tool_bar {
        height: 3;
        background: $surface-darken-1;
        #max_lines {
          width: 10;
        }
        Label {
          margin-top: 1;
          background: transparent;
        }
      }
      #logs {
        border: solid $primary;
      }
    }
    """

    richlog: RichLog
    max_lines_input: Input

    def __init__(self, **kwargs) -> None:
        """Initialise the screen."""
        super().__init__(**kwargs)
        self.richlog = RichLog(id="logs", wrap=True, highlight=True, auto_scroll=True)
        self.auto_scroll = Checkbox(label="Auto Scroll", value=True, id="auto_scroll")
        self.max_lines_input = Input(
            id="max_lines",
            placeholder="Max Lines",
            value=str(settings.max_log_lines),
            type="integer",
        )
        self.richlog.max_lines = int(self.max_lines_input.value)

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""

        with Vertical(id="menu"):
            with Horizontal(id="tool_bar"):
                yield self.auto_scroll
                yield Label("Max Lines: ")
                yield self.max_lines_input
                yield Button("Clear", id="clear", variant="warning")
            yield self.richlog

        self.richlog.write("Starting...")

    def _on_show(self, event: Show) -> None:
        """Handle show event"""
        self.screen.sub_title = "Logs"  # pylint: disable=attribute-defined-outside-init

    @on(Checkbox.Changed, "#auto_scroll")
    def on_auto_scroll_changed(self, event: Checkbox.Changed) -> None:
        """Handle auto scroll checkbox change"""
        self.richlog.auto_scroll = event.value

    @on(Button.Pressed, "#clear")
    def on_clear_pressed(self) -> None:
        """Handle clear button press"""
        self.richlog.clear()

    @on(Input.Changed, "#max_lines")
    def on_max_lines_changed(self, event: Input.Changed) -> None:
        """Handle max lines input change"""
        max_lines: int = max(0, int(event.value or "0"))
        self.richlog.max_lines = max_lines
        settings.max_log_lines = max_lines
        settings.save_settings_to_file()



"""Field set consisting of Label and Input for editing items."""
from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import Button
from textual.widgets import Input
from textual.widgets import Label
from textual.widgets import Select
from textual.widgets import TextArea

from parllama.icons import COPY_EMOJI
from parllama.messages.main import SendToClipboard
from parllama.utils import to_class_case
from parllama.widgets.hidden_input import HiddenInputField


class FieldSet(Widget):
    """Field set consisting of Label and Input for editing items."""

    DEFAULT_CSS = """
    FieldSet {
        width: 1fr;
        height: 1;
        Label {
            width: 9;
            margin-right: 1;
        }
        &> Horizontal {
            width: 1fr;
            &> Horizontal {
                width: 3fr;
            }
        }

        Input, Select, TextArea, Checkbox, HiddenInputField{
            width: 1fr;
        }
    }
    """
    label: Label
    """Label widget."""
    input: Widget
    """Input widget."""

    show_copy_button: bool = False
    """Show copy button to right of input."""
    _extra_children: list[Widget] | None = None

    def __init__(
        self,
        label: str,
        input_widget: Widget,
        *,
        show_copy_button: bool = False,
        extra_children: list[Widget] | None = None,
    ) -> None:
        """Initialize the field set."""
        base_name = to_class_case(label.replace("/", ""))
        super().__init__(
            id=f"{base_name}FieldSet", name=f"{base_name}FieldSet", classes="field_set"
        )
        self.show_copy_button = show_copy_button
        self._extra_children = extra_children
        self.input = input_widget
        self.label = Label(label, id=f"{base_name}Label")

    def compose(self) -> ComposeResult:
        """Compose the field set."""
        with Horizontal():
            yield self.label
            with Horizontal():
                yield self.input
                if self.show_copy_button:
                    btn = Button(COPY_EMOJI, id="copy", classes="field-button")
                    btn.tooltip = "Copy to clipboard"
                    yield btn
                if self._extra_children:
                    yield from self._extra_children

    @on(Button.Pressed, "#copy")
    def copy(self, event: Button.Pressed) -> None:
        """Copy field value."""
        event.stop()

        if isinstance(self.input, (Input, Select)):
            self.app.post_message(
                SendToClipboard(str(self.input.value) if self.input.value else "")
            )
        elif isinstance(self.input, TextArea):
            self.app.post_message(SendToClipboard(self.input.text))
        elif isinstance(self.input, HiddenInputField):
            self.app.post_message(SendToClipboard(self.input.value))



"""Chat manager class"""
from __future__ import annotations

import datetime
import os
import uuid
from collections.abc import Iterator
from collections.abc import Mapping
from dataclasses import dataclass
from io import StringIO
from typing import Any
from typing import Literal

import simplejson as json
from ollama import Message as OMessage
from ollama import Options as OllamaOptions
from textual.widget import Widget

from parllama.messages.main import ChatMessage
from parllama.models.settings_data import settings


@dataclass
class OllamaMessage:
    """
    Chat message.
    """

    message_id: str
    "Unique identifier of the message."

    role: Literal["user", "assistant", "system"]
    "Assumed role of the message. Response messages always has role 'assistant'."

    content: str = ""
    "Content of the message. Response messages contains message fragments when streaming."

    session: ChatSession | None = None

    def __init__(
        self,
        *,
        role: Literal["user", "assistant", "system"],
        content: str,
        message_id: str | None = None,
        session: ChatSession | None = None,
    ) -> None:
        """Initialize the chat message"""
        self.message_id = message_id or uuid.uuid4().hex
        self.role = role
        self.content = content
        self.session = session
        self.id_to_msg = {self.message_id: self}

    def __str__(self) -> str:
        """Ollama message representation"""
        return f"## {self.role}\n\n{self.content}\n\n"

    def to_ollama_native(self) -> OMessage:
        """Convert a message to Ollama native format"""
        return OMessage(role=self.role, content=self.content)

    def save(self) -> None:
        """Save the chat session to a file"""
        if self.session:
            self.session.save()

    def to_json(self) -> str:
        """Convert the chat session to JSON"""
        return json.dumps(
            {"message_id": self.message_id, "role": self.role, "content": self.content},
            default=str,
            indent=4,
        )

    def __dict__(
        self,
    ):
        """Convert the chat message to a dictionary"""
        return {
            "message_id": self.message_id,
            "role": self.role,
            "content": self.content,
        }

    @staticmethod
    def from_json(json_data: str) -> OllamaMessage:
        """Convert JSON to chat session"""
        data: dict = json.loads(json_data)
        return OllamaMessage(
            message_id=data["message_id"], role=data["role"], content=data["content"]
        )


@dataclass
class ChatSession:
    """Chat session class"""

    session_id: str
    llm_model_name: str
    options: OllamaOptions
    session_name: str
    messages: list[OllamaMessage]
    id_to_msg: dict[str, OllamaMessage]
    last_updated: datetime.datetime

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        *,
        session_id: str | None = None,
        llm_model_name: str,
        options: OllamaOptions | None = None,
        session_name: str,
        messages: list[OllamaMessage] | list[dict] | None = None,
        last_updated: datetime.datetime | None = None,
    ):
        """Initialize the chat session"""
        self.session_id = session_id or uuid.uuid4().hex
        self.llm_model_name = llm_model_name
        self.options = options or {}
        self.session_name = session_name
        self.messages = []
        msgs = messages or []
        for m in msgs:
            if isinstance(m, OllamaMessage):
                self.messages.append(m)
            else:
                self.messages.append(OllamaMessage(session=self, **m))

        self.id_to_msg = {}
        self.last_updated = last_updated or datetime.datetime.now()

        for m in self.messages:
            m.session = self
            self.id_to_msg[m.message_id] = m

    def get_message(self, message_id: str) -> OllamaMessage | None:
        """Get a message"""
        return self.id_to_msg.get(message_id)

    def add_message(self, msg: OllamaMessage) -> None:
        """Add a message"""
        msg.session = self
        self.messages.append(msg)
        self.id_to_msg[msg.message_id] = msg
        self.last_updated = datetime.datetime.now()
        self.save()

    def set_name(self, name: str) -> None:
        """Set the name of the chat session"""
        self.session_name = name
        self.last_updated = datetime.datetime.now()
        self.save()

    def set_llm_model(self, llm_model_name: str) -> None:
        """Set the LLM model name"""
        self.llm_model_name = llm_model_name
        self.last_updated = datetime.datetime.now()
        self.save()

    def set_temperature(self, temperature: float | None) -> None:
        """Set the temperature"""
        if temperature:
            self.options["temperature"] = temperature
        else:
            del self.options["temperature"]
        self.last_updated = datetime.datetime.now()
        self.save()

    async def send_chat(self, from_user: str, widget: Widget) -> bool:
        """Send a chat message to LLM"""
        msg: OllamaMessage = OllamaMessage(content=from_user, role="user")
        self.add_message(msg)
        widget.post_message(
            ChatMessage(session_id=self.session_id, message_id=msg.message_id)
        )

        msg = OllamaMessage(content="", role="assistant")
        self.add_message(msg)
        widget.post_message(
            ChatMessage(session_id=self.session_id, message_id=msg.message_id)
        )

        stream: Iterator[Mapping[str, Any]] = settings.ollama_client.chat(  # type: ignore
            model=self.llm_model_name,
            messages=[m.to_ollama_native() for m in self.messages],
            options=self.options,
            stream=True,
        )

        for chunk in stream:
            msg.content += chunk["message"]["content"]
            widget.post_message(
                ChatMessage(session_id=self.session_id, message_id=msg.message_id)
            )
        msg.save()
        return True

    def new_session(self, session_name: str = "My Chat"):
        """Start new session"""
        self.session_id = uuid.uuid4().hex
        self.session_name = session_name
        self.messages.clear()
        self.id_to_msg.clear()

    def __iter__(self):
        """Iterate over messages"""
        return iter(self.messages)

    def __len__(self) -> int:
        """Get the number of messages"""
        return len(self.messages)

    def __getitem__(self, msg_id: str) -> OllamaMessage:
        """Get a message"""
        return self.id_to_msg[msg_id]

    def __setitem__(self, msg_id: str, value: OllamaMessage) -> None:
        """Set a message"""
        self.id_to_msg[msg_id] = value
        for i, msg in enumerate(self.messages):
            if msg.message_id == msg_id:
                self.messages[i] = value
                return
        self.messages.append(value)

    def __delitem__(self, key: str) -> None:
        """Delete a message"""
        del self.id_to_msg[key]
        for i, msg in enumerate(self.messages):
            if msg.message_id == key:
                self.messages.pop(i)
                return

    def __contains__(self, item: OllamaMessage) -> bool:
        """Check if a message exists"""
        return item.message_id in self.id_to_msg

    def __eq__(self, other: object) -> bool:
        """Check if two sessions are equal"""
        if not isinstance(other, ChatSession):
            return NotImplemented
        return self.session_id == other.session_id

    def __ne__(self, other: object) -> bool:
        """Check if two sessions are not equal"""
        if not isinstance(other, ChatSession):
            return NotImplemented
        return self.session_id != other.session_id

    def __str__(self) -> str:
        """Get a string representation of the chat session"""
        ret = StringIO()
        ret.write(f"# {self.session_name}\n\n")
        for msg in self.messages:
            ret.write(str(msg))
        return ret.getvalue()

    def to_json(self) -> str:
        """Convert the chat session to JSON"""
        return json.dumps(
            {
                "session_id": self.session_id,
                "last_updated": self.last_updated.isoformat(),
                "llm_model_name": self.llm_model_name,
                "options": self.options,
                "session_name": self.session_name,
                "messages": [m.__dict__() for m in self.messages],
            },
            default=str,
            indent=4,
        )

    @staticmethod
    def from_json(json_data: str) -> ChatSession:
        """Convert JSON to chat session"""
        data: dict = json.loads(json_data)
        return ChatSession(
            session_id=data["session_id"],
            last_updated=datetime.datetime.fromisoformat(data["last_updated"]),
            llm_model_name=data["llm_model_name"],
            options=data.get("options"),
            session_name=data["session_name"],
            messages=[OllamaMessage(**m) for m in data["messages"]],
        )

    @staticmethod
    def load_from_file(filename: str) -> ChatSession | None:
        """Load a chat session from a file"""
        try:
            with open(
                os.path.join(settings.chat_dir, filename), "rt", encoding="utf-8"
            ) as f:
                return ChatSession.from_json(f.read())
        except (OSError, IOError):
            return None

    def is_valid(self) -> bool:
        """return true if session is valid"""
        return (
            len(self.session_name) > 0
            and len(self.llm_model_name) > 0
            and self.llm_model_name not in ["Select.BLANK", "None", ""]
            and len(self.messages) > 0
        )

    def save(self) -> bool:
        """Save the chat session to a file"""
        if not self.is_valid():
            return False  # Cannot save without session name, LLM model name and at least one message

        file_name = (
            f"{self.session_id}.json"  # Use session ID as filename to avoid over
        )
        try:
            with open(
                os.path.join(settings.chat_dir, file_name), "wt", encoding="utf-8"
            ) as f:
                f.write(self.to_json())
            return True
        except (OSError, IOError):
            return False

    def export_as_markdown(self, filename: str) -> bool:
        """Save the chat session to markdown file"""
        try:
            with open(
                os.path.join(settings.chat_dir, filename), "w", encoding="utf-8"
            ) as f:
                f.write(str(self))
            return True
        except (OSError, IOError):
            return False



"""Chat message widget"""
from __future__ import annotations

from textual.await_complete import AwaitComplete
from textual.binding import Binding
from textual.widgets import Markdown

from parllama.messages.main import SendToClipboard
from parllama.models.chat import OllamaMessage


class ChatMessageWidget(Markdown, can_focus=True):
    """Chat message widget base"""

    BINDINGS = [
        Binding("ctrl+c", "copy_to_clipboard", "", show=True),
    ]

    DEFAULT_CSS = """
    ChatMessageWidget {
        margin: 0;
        MarkdownFence {
            margin: 1 2;
            max-height: initial;
        }
    }
    """
    msg: OllamaMessage

    def __init__(self, msg: OllamaMessage, **kwargs) -> None:
        """Initialise the widget."""
        super().__init__(**kwargs)
        self.msg = msg
        self.update("")

    def update(self, markdown: str) -> AwaitComplete:
        """Update the document with new Markdown."""
        self.msg.content += markdown
        return super().update("## " + self.msg.role + "\n\n" + self.msg.content)

    @property
    def raw_text(self) -> str:
        """The raw text."""
        return self.msg.content or ""

    @staticmethod
    def mk_msg_widget(msg: OllamaMessage) -> ChatMessageWidget:
        """Create a chat message widget."""
        if msg.role == "user":
            return UserChatMessage(msg=msg)
        return AgentChatMessage(msg=msg)

    def action_copy_to_clipboard(self) -> None:
        """Copy focused widget value to clipboard"""
        self.app.post_message(SendToClipboard(self.raw_text))


class AgentChatMessage(ChatMessageWidget):
    """Agent chat message widget"""

    DEFAULT_CSS = """
    AgentChatMessage {
      background: $panel-lighten-2;
    }
    AgentChatMessage:light {
        background: #ccc;
    }
    """


class UserChatMessage(ChatMessageWidget):
    """User chat message widget"""

    DEFAULT_CSS = """
    UserChatMessage {
       background: $surface;
    }
    UserChatMessage:light {
        background: #aaa;
    }
    """



"""Various utility functions and decorators."""
from __future__ import annotations

import csv
import hashlib
import io
import math
import os
import random
import string
import subprocess
import sys
import time
import uuid
from argparse import ArgumentParser
from argparse import Namespace
from collections.abc import Generator
from contextlib import contextmanager
from datetime import date
from datetime import datetime
from decimal import Decimal
from io import StringIO
from os import listdir
from os.path import isfile
from os.path import join
from typing import Any
from typing import Literal
from typing import TypeAlias

from textual.widgets import Button
from textual.widgets import Input
from textual.widgets.button import ButtonVariant

from parllama import __application_binary__
from parllama import __application_title__
from parllama import __version__
from parllama.icons import PENCIL_EMOJI
from parllama.icons import TRASH_EMOJI

DECIMAL_PRECESSION = 5

ScreenType: TypeAlias = Literal["Local", "Site", "Tools", "Create", "Chat", "Logs"]
valid_screens: list[ScreenType] = ["Local", "Site", "Tools", "Create", "Chat", "Logs"]


def id_generator(
    size: int = 6, chars: str = string.ascii_uppercase + string.digits
) -> str:
    """
    Generates a random string of uppercase letters and digits.

    :param size: The length of the string to generate.
    :param chars: The characters to use for the string.
    :return: The random string.
    """
    return "".join(random.choice(chars) for _ in range(size))


def json_serial(obj: Any) -> str:
    """
    JSON serializer for objects not serializable by default json code.

    :param obj: The object to serialize.
    :return: The serialized object.
    """

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError(f"Type {type(obj)} not serializable")


def coalesce(*arg):  # type: ignore
    """
    Return first item that is not None.

    :param arg: The items to check.
    :return: The first non-None item.
    """
    return next((a for a in arg if a is not None), None)


def chunks(lst: list, n: int) -> Generator[list, None, None]:
    """
    Yield successive n-sized chunks from lst.

    :param lst: The list to split.
    :param n: The size of the chunks.
    :return: The chunks.
    """
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def to_camel_case(snake_str: str) -> str:
    """
    Convert a snake_case string to CamelCase.

    :param snake_str: The snake_case string.
    :return: The CamelCase string.
    """
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + "".join(x.title() for x in components[1:])


def to_class_case(snake_str: str) -> str:
    """
    Convert a snake_case string to ClassCase.
    Spaces are converted to underscores before conversion.

    :param snake_str: The snake_case string.
    :return: The ClassCase string.
    """
    components = snake_str.replace(" ", "_").split("_")
    # We capitalize the first letter of each component
    # with the 'title' method and join them together.
    return "".join(x.title() for x in components[0:])


def get_files(path: str, ext: str = "") -> list[str]:
    """Return list of file names in alphabetical order inside of provided path non-recursively.
    Omitting files not ending with ext."""
    ret = [
        f
        for f in listdir(path)
        if isfile(join(path, f)) and (not ext or not f.endswith(ext))
    ]
    ret.sort()
    return ret


# tests if value is able to be converted to float
def is_float(s: any) -> bool:  # type: ignore
    """Test if value is able to be converted to float."""
    try:
        float(s)
        return True
    except (ValueError, TypeError):
        return False


# tests if value is able to be converted to int
def is_int(s: any) -> bool:  # type: ignore
    """Test if value is able to be converted to int."""
    try:
        int(s)
        return True
    except (ValueError, TypeError):
        return False


def is_date(date_text: str, fmt: str = "%Y/%m/%d") -> bool:
    """Test if value is a valid date in format of provided format defaults to "%Y/%m/%d"."""
    try:
        datetime.strptime(date_text, fmt)
        return True
    except ValueError:
        return False


# pylint: disable=too-many-return-statements, too-many-branches
def has_value(v: Any, search: str, depth: int = 0) -> bool:
    """Recursively search data structure for search value"""
    # don't go more than 3 levels deep
    if depth > 4:
        return False
    # if is a dict, search all dict values recursively
    if isinstance(v, dict):
        for dv in v.values():
            if has_value(dv, search, depth + 1):
                return True
    # if is a list, search all list values recursively
    if isinstance(v, list):
        for li in v:
            if has_value(li, search, depth + 1):
                return True
    # if is an int, trim off .00 for search if it exists then compare
    if isinstance(v, int):
        search = search.rstrip(".00")
        if str(v) == search:
            return True
    # if is a float, truncate string version of float to same size as search
    if isinstance(v, float):
        v = str(v)[0 : len(search)]
        if search == v:
            return True
    # if is a string, strip and lowercase it then check if string starts with search
    if isinstance(v, str):
        if v.strip().lower().startswith(search) or v.strip().lower().endswith(search):
            return True
    return False


def is_zero(val: any) -> bool:  # type: ignore
    """Test if value is zero."""
    if val is None:
        return False
    t = type(val)
    if t is Decimal:
        return val.round(DECIMAL_PRECESSION).is_zero()
    if t is float:
        return math.isclose(round(val, 5), 0, rel_tol=1e-05)
    if t is int:
        return 0 == val
    return False


def non_zero(val: any) -> bool:  # type: ignore
    """Test if value is not zero."""
    return not is_zero(val)


def dict_keys_to_lower(dictionary: dict) -> dict:
    """
    Return a new dictionary with all keys lowercase
    @param dictionary: dict with keys that you want to lowercase
    @return: new dictionary with lowercase keys
    """
    return {k.lower(): v for k, v in dictionary.items()}


def is_valid_uuid_v4(value: str) -> bool:
    """Test if value is a valid UUID v4."""
    try:
        uuid_obj = uuid.UUID(value, version=4)
        return str(uuid_obj) == value  # Check if the string representation matches
    except ValueError:
        return False


def parse_csv_text(csv_data: StringIO) -> list[dict]:
    """
    Reads in a CSV file as text and returns it as a list of dictionaries.

    Args:
            csv_data (StringIO): The CSV file as text.

    Returns:
            list[dict]: The CSV data as a list of dictionaries.
    """
    return list(csv.DictReader(csv_data))


def read_text_file_to_stringio(file_path: str, encoding: str = "utf-8") -> StringIO:
    """
    Reads in a text file and returns it as a StringIO object.

    Args:
            file_path (str): The path to the file to read.
            encoding (str): The encoding of the file.

    Returns:
            StringIO: The text file as a StringIO object.
    """
    with open(file_path, encoding=encoding) as file:
        return StringIO(file.read())


def md5_hash(data: str) -> str:
    """
    Returns a md5 hash of the input data.

    Args:
            data (str): The input data.

    Returns:
            str: The md5 hash of the input data.
    """
    md5 = hashlib.md5()
    md5.update(data.encode())
    return md5.hexdigest()


def sha1_hash(data: str) -> str:
    """
    Returns a SHA1 hash of the input data.

    Args:
            data (str): The input data.

    Returns:
            str: The SHA1 hash of the input data.
    """
    sha1 = hashlib.sha1()
    sha1.update(data.encode())
    return sha1.hexdigest()


def sha256_hash(data: str) -> str:
    """
    Returns a SHA256 hash of the input data.

    Args:
            data (str): The input data.

    Returns:
            str: The SHA256 hash of the input data.
    """
    sha256 = hashlib.sha256()
    sha256.update(data.encode())
    return sha256.hexdigest()


def nested_get(dictionary: dict, keys: str | list[str]) -> Any:
    """
    Returns the value for a given key in a nested dictionary.

    Args:
            dictionary (dict): The nested dictionary to search.
            keys (str | list[str]): The key or list of keys to search for.

    Returns:
            Any: The value for the given key or None if the key does not exist.
    """
    if isinstance(keys, str):
        keys = keys.split(".")
    if keys and dictionary:
        element = keys[0]
        if element in dictionary:
            if len(keys) == 1:
                return dictionary[element]
            return nested_get(dictionary[element], keys[1:])
    return None


def override(cls):  # type: ignore
    """
    A decorator to ensure that the decorated method overrides a method in its superclass.
    """

    def overrider(method):  # type: ignore
        if method.__name__ not in dir(cls):
            raise AttributeError(
                f"Method {method.__name__} is not overriding any method of {cls.__name__}."
            )
        return method

    return overrider


@contextmanager
def add_module_path(path: str) -> Generator[None, None, None]:
    """Add a module path to sys.path temporarily."""
    sys.path.append(path)
    try:
        yield
    finally:
        try:
            sys.path.remove(path)
        except ValueError:
            # path is not in sys.path
            pass


@contextmanager
def catch_to_logger(logger: any, re_throw: bool = False):  # type: ignore
    """Catch exceptions and log them to a logger."""
    try:
        yield
    # pylint: disable=broad-except
    except Exception as e:
        if logger and hasattr(logger, "exception"):
            logger.exception(e)  # type: ignore
            if re_throw:
                raise e
        else:
            raise e


@contextmanager
def timer_block(label: str = "Timer") -> Generator[None, None, None]:
    """Time a block of code."""

    start_time = time.time()
    try:
        yield
    finally:
        end_time = time.time()
        elapsed_time = end_time - start_time
        print(f"{label} took {elapsed_time:.4f} seconds.")


def str_ellipsis(s: str, max_len: int) -> str:
    """Return a left space padded string exactly max_len with ellipsis if it exceeds max_len."""
    if len(s) <= max_len:
        return s.ljust(max_len)
    return s[: max_len - 3] + "..."


def detect_syntax(text: str) -> str | None:
    """Detect the syntax of the text."""
    lines = text.split("\n")
    if len(lines) > 0:
        line = lines[0]
        if line.startswith("#!"):
            if (
                line.endswith("/bash")
                or line.endswith("/sh")
                or line.endswith(" bash")
                or line.endswith(" sh")
            ):
                return "bash"
    return None


def mk_field_button(
    *,
    id: str,  # pylint: disable=redefined-builtin
    classes: str = "",
    emoji: str,
    tooltip: str = "",
    variant: ButtonVariant = "default",
) -> Button:
    """Make a field button."""
    btn = Button(emoji, id=id, classes=f"field-button {classes}", variant=variant)
    if tooltip:
        btn.tooltip = tooltip
    return btn


def mk_trash_button(
    *,
    id: str = "delete",  # pylint: disable=redefined-builtin
    classes: str = "",
    tooltip: str = "Delete",
) -> Button:
    """Make a trash button."""
    return mk_field_button(
        id=id,
        classes=classes,
        emoji=TRASH_EMOJI,
        tooltip=tooltip,
        variant="error",
    )


def mk_edit_button(
    *,
    id: str = "edit",  # pylint: disable=redefined-builtin
    classes: str = "",
    tooltip: str = "Edit",
) -> Button:
    """Make a edit button."""
    return mk_field_button(
        id=id,
        classes=classes,
        emoji=PENCIL_EMOJI,
        tooltip=tooltip,
    )


def hash_list_by_key(data: list[dict], id_key: str = "message_id") -> dict:
    """Hash a list of dictionaries by a key."""
    return {item[id_key]: item for item in data}


def run_shell_cmd(cmd: str) -> str | None:
    """Run a command and return the output."""
    try:
        return subprocess.run(
            cmd, shell=True, capture_output=True, check=True, encoding="utf-8"
        ).stdout.strip()
    except:  # pylint: disable=bare-except
        return None


def clamp_input_value(
    input_widget: Input, min_value: int | None = None, max_value: int | None = None
) -> int:
    """Clamp the value of an input widget."""
    val: int = int(input_widget.value or 0)
    if min_value is not None and val < min_value:
        input_widget.value = str(min_value)
        return min_value
    if max_value is not None and val > max_value:
        input_widget.value = str(max_value)
        return max_value
    return val


eff_word_list_cache: list[str] = []


def get_args() -> Namespace:
    """Parse and return the command line arguments.

    Returns:
            The result of parsing the arguments.
    """

    # Create the parser object.
    parser = ArgumentParser(
        prog=__application_binary__,
        description=f"{__application_title__} -- Ollama TUI.",
        epilog=f"v{__version__}",
    )

    # Add --version
    parser.add_argument(
        "-v",
        "--version",
        help="Show version information.",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    parser.add_argument(
        "-d",
        "--data-dir",
        help="Data Directory. Defaults to ~/.parllama",
    )

    parser.add_argument(
        "-u",
        "--ollama-url",
        help="URL of your Ollama instance. Defaults to http://localhost:11434",
    )

    parser.add_argument("-t", "--theme-name", help="Theme name. Defaults to par")
    parser.add_argument(
        "-m",
        "--theme-mode",
        help="Dark / Light mode. Defaults to dark",
        choices=["dark", "light"],
    )

    parser.add_argument(
        "-s",
        "--starting-screen",
        help="Starting screen. Defaults to local",
        choices=[s.lower() for s in valid_screens],
    )

    parser.add_argument(
        "-p",
        "--ps-poll",
        type=int,
        help="Interval in seconds to poll ollama ps command. 0 = disable. Defaults to 3",
    )

    parser.add_argument(
        "--restore-defaults",
        help="Restore default settings and theme",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--clear-cache",
        help="Clear cached data",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--purge-chats",
        help="Purge all chat history",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--no-save",
        help="Prevent saving settings for this session.",
        default=False,
        action="store_true",
    )

    # Finally, parse the command line.
    return parser.parse_args()


def output_to_dicts(output: str) -> list[dict[str, Any]]:
    """Convert a tab-delimited output to a list of dicts."""
    if not output:
        return []
    # split string on newline loop over each line and convert
    # Use csv module to parse the tab-delimited output
    reader = csv.DictReader(io.StringIO(output), delimiter="\t")
    ret = []
    for model in reader:
        mod = {}
        for key, value in model.items():
            mod[key.strip().lower()] = value.strip()
        ret.append(mod)
    return ret


def run_cmd(params: list[str]) -> str | None:
    """Run a command and return the output."""
    try:
        result = subprocess.run(
            params, capture_output=True, text=True, check=True, universal_newlines=True
        )
        ret = result.stdout.strip()
        # Split the output into lines
        lines = [
            line
            for line in ret.splitlines()
            if not line.startswith("failed to get console mode")
        ]

        # Get the last two lines
        return "\n".join(lines)
    except subprocess.CalledProcessError as e:
        print(f"Error running command {e.stderr}")
        return None


def read_env_file(filename: str) -> dict[str, str]:
    """
    Read environment variables from a file into a dictionary

    Args:
        filename (str): The name of the file to read

    Returns:
        Dict[str, str]: A dictionary containing the environment variables
    """
    env_vars: dict[str, str] = {}
    if not os.path.exists(filename):
        return env_vars
    with open(filename, encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#") or not "=" in line:
                continue
            try:
                key, value = line.split("=", 1)
                env_vars[key.strip()] = value.strip()
            except Exception as e:  # pylint: disable=broad-exception-caught
                print(f"Error: {e} --- line {line}")
    return env_vars



"""Provides model details dialog."""
from __future__ import annotations

import re
from typing import cast

import humanize
import ollama
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import VerticalScroll
from textual.events import Focus
from textual.screen import ModalScreen
from textual.widgets import Button
from textual.widgets import MarkdownViewer
from textual.widgets import Static
from textual.widgets import TextArea

from ..messages.main import CreateModelFromExistingRequested
from ..models.ollama_data import FullModel
from ..models.ollama_data import MessageRoles
from ..widgets.field_set import FieldSet


class ModelDetailsDialog(ModalScreen[None]):
    """Modal dialog that shows model details."""

    DEFAULT_CSS = """
    ModelDetailsDialog {
        background: black 75%;
        align: center middle;
        &> VerticalScroll {
            background: $surface;
            width: 75%;
            height: 90%;
            min-width: 80;
            border: thick $accent;
            border-title-color: $primary;
            padding: 1;
        }
        .editor {
            background: $panel;
            border: solid $primary;
            border-title-color: $primary;
            margin-bottom: 1;
            .text-area--cursor-line {
              background: transparent;
            }
        }
    }
    """

    BINDINGS = [
        Binding("left,up", "app.focus_previous", "", show=False),
        Binding("right,down", "app.focus_next", "", show=False),
        Binding("escape, ctrl+q", "app.pop_screen", "", show=True),
        Binding("ctrl+c", "app.copy_to_clipboard", "", show=True),
    ]
    model: FullModel

    def __init__(self, model: FullModel) -> None:
        super().__init__()
        self.model = model

    def compose(self) -> ComposeResult:
        """Compose the content of the dialog."""
        with VerticalScroll() as vs:
            vs.border_title = f"[ {self.model.name} ]"
            yield Button("Copy to create", id="copy_to_create")
            # yield FieldSet("Name", Static(self.model.name, message_id="name"))
            yield FieldSet(
                "Modified", Static(str(self.model.modified_at), id="modified_at")
            )
            exp = str(self.model.expires_at)
            if exp == "0001-01-01 00:00:00+00:00":
                exp = "Never"
            yield FieldSet("Expires", Static(exp, id="expires_at"))
            yield FieldSet(
                "Size", Static(humanize.naturalsize(self.model.size), id="size")
            )
            yield FieldSet("Digest", Static(self.model.digest, id="digest"))
            yield Static("")
            ta = TextArea(
                self.model.template or "", id="template", classes="editor height-auto"
            )
            ta.border_title = "Template"
            ta.read_only = True
            yield ta

            ta = TextArea(
                self.model.parameters or "",
                id="parameters",
                classes="editor height-auto",
            )
            ta.border_title = "Parameters"
            ta.read_only = True
            yield ta

            ta = TextArea(
                self.model.modelfile, id="modelfile", classes="editor height-10"
            )
            ta.border_title = "Model file"
            ta.read_only = True
            yield ta

            ta = TextArea(
                self.model.license or "", id="license", classes="editor height-10"
            )
            ta.border_title = "License"
            ta.read_only = True
            yield ta

            system_regex = re.compile(r"^system (.*)", re.I)
            message_regex = re.compile(r"^message (user|assistant|system) (.*)", re.I)
            messages: list[ollama.Message] = []
            system_msg: str = ""
            for line in self.model.modelfile.splitlines():
                match = message_regex.match(line)
                if match:
                    messages.append(
                        ollama.Message(
                            role=cast(MessageRoles, match.group(1)),
                            content=match.group(2),
                        )
                    )
                match = system_regex.match(line)
                if match:
                    system_msg = match.group(1)

            msgs = [f"* MESSAGE {m['role']} {m['content']}" for m in messages]
            if system_msg:
                msgs.insert(0, f"* SYSTEM {system_msg}")
            md = MarkdownViewer(
                "\n".join(msgs),
                id="messages",
                classes="editor height-auto",
                show_table_of_contents=False,
            )
            md.border_title = "Messages"
            yield md

    async def on_mount(self) -> None:
        """Mount the view."""
        self.query_one("#copy_to_create").focus()

    @on(Button.Pressed, "#copy_to_create")
    def copy_to_create(self, event: Button.Pressed) -> None:
        """Copy model to create screen."""
        event.stop()
        with self.prevent(Focus):
            self.app.pop_screen()
        self.app.post_message(
            CreateModelFromExistingRequested(
                widget=None,
                model_name=self.model.name,
                model_code=self.model.modelfile,
                quantization_level=None,
            )
        )



"""Provides a dialog for getting a yes/no response from the user."""
from __future__ import annotations

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center
from textual.containers import Horizontal
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Button
from textual.widgets import Static


class YesNoDialog(ModalScreen[bool]):
    """A dialog for asking a user a yes/no question."""

    DEFAULT_CSS = """
	YesNoDialog {
		align: center middle;
	}

	YesNoDialog > Vertical {
		background: $panel;
		height: auto;
		width: auto;
		border: thick $secondary;
	}

	YesNoDialog > Vertical > * {
		width: auto;
		height: auto;
	}

	YesNoDialog Static {
		width: auto;
	}

	YesNoDialog .spaced {
		padding: 1;
	}

	YesNoDialog #question {
		min-width: 100%;
		border-top: solid $secondary;
		border-bottom: solid $secondary;
	}

	YesNoDialog Button {
		margin-right: 1;
	}

	YesNoDialog #buttons {
		width: 100%;
		align-horizontal: right;
		padding-right: 1;
	}
	"""

    BINDINGS = [
        Binding("left,up", "app.focus_previous", "", show=False),
        Binding("right,down", "app.focus_next", "", show=False),
        Binding("escape", "dismiss(False)", "", show=False),
    ]

    def __init__(  # pylint:disable=too-many-arguments
        self,
        title: str,
        question: str,
        yes_label: str = "Yes",
        no_label: str = "No",
        yes_first: bool = True,
    ) -> None:
        """Initialise the yes/no dialog.

        Args:
                title: The title for the dialog.
                question: The question to ask.
                yes_label: The optional label for the yes button.
                no_label: The optional label for the no button.
                yes_first: Should the yes button come first?
        """
        super().__init__()
        self._title = title
        self._question = question
        self._yes = yes_label
        self._no = no_label
        self._yes_first = yes_first

    def compose(self) -> ComposeResult:
        """Compose the content of the dialog."""
        with Vertical():
            with Center():
                yield Static(self._title, classes="spaced")
            yield Static(self._question, id="question", classes="spaced")
            with Horizontal(id="buttons"):
                aye = Button(self._yes, id="yes")
                naw = Button(self._no, id="no")
                if self._yes_first:
                    aye.variant = "primary"
                    yield aye
                    yield naw
                else:
                    naw.variant = "primary"
                    yield naw
                    yield aye

    def on_mount(self) -> None:
        """Configure the dialog once the DOM is ready."""
        self.query(Button).first().focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle a button being pressed on the dialog."""
        self.dismiss(event.button.id == "yes")



"""Theme manager for Textual"""
from __future__ import annotations

import os
import shutil
from pathlib import Path
from typing import Dict
from typing import Literal
from typing import TypeAlias

import simplejson as json
from textual.design import ColorSystem

from parllama.models.settings_data import settings

ThemeMode: TypeAlias = Literal["dark", "light"]
ThemeModes: list[ThemeMode] = ["dark", "light"]
Theme: TypeAlias = Dict[ThemeMode, ColorSystem]
Themes: TypeAlias = Dict[str, Theme]


class InvalidThemeError(Exception):
    """Raised when an invalid theme is provided."""

    def __init__(self, theme_name: str):
        """Initialize the exception with the invalid theme name."""
        self.theme_name = theme_name
        super().__init__(f"Invalid theme: {theme_name}")


class ThemeModeError(InvalidThemeError):
    """Raised when a theme does not have at least one of 'dark' or 'light' modes."""

    def __init__(self, theme_name: str):
        """Initialize the exception with the invalid theme name."""
        super().__init__(
            f"Theme '{theme_name}' does not have at least one of 'dark' or 'light' modes."
        )


class ThemeManager:
    """Theme manager for Textual"""

    themes: Themes
    theme_folder: str

    def __init__(self) -> None:
        """Initialize the theme manager"""
        self.theme_folder = os.path.join(settings.data_dir, "themes")

        self.themes = self.load_themes()

    def ensure_default_theme(self) -> None:
        """Ensure that the default theme exists."""
        default_theme_file: str = os.path.join(
            os.path.dirname(Path(__file__)), "themes", "par.json"
        )
        if not os.path.exists(self.theme_folder):
            os.makedirs(self.theme_folder)
        theme_file = os.path.join(self.theme_folder, "par.json")

        if not os.path.exists(theme_file):
            shutil.copy(default_theme_file, theme_file)

    def load_theme(self, theme_name: str) -> Theme:
        """Load textual theme from json file"""

        theme: Theme = {}
        theme_name = os.path.basename(theme_name)
        with open(os.path.join(self.theme_folder, theme_name), encoding="utf-8") as f:
            theme_def = json.load(f)
            if "dark" not in theme_def and "light" not in theme_def:
                raise ThemeModeError(theme_name)

            for mode in ThemeModes:
                if mode in theme_def:
                    theme[mode] = ColorSystem(**theme_def[mode])
        return theme

    def load_themes(self) -> Themes:
        """Load textual themes from json files"""
        self.ensure_default_theme()

        themes: Themes = {}
        for file in os.listdir(self.theme_folder):
            if file.lower().endswith(".json"):
                theme_name = os.path.splitext(file)[0]
                themes[theme_name] = self.load_theme(f"{self.theme_folder}/{file}")
        return themes

    def get_theme(self, theme_name: str) -> Theme:
        """Get theme by name"""
        return self.themes[theme_name]

    def list_themes(self) -> list[str]:
        """Get list of themes"""
        return list(self.themes.keys())

    def theme_has_dark(self, theme_name: str) -> bool:
        """Check if theme has dark mode"""
        return "dark" in self.themes[theme_name]

    def theme_has_light(self, theme_name: str) -> bool:
        """Check if theme has light mode"""
        return "light" in self.themes[theme_name]

    def get_color_system_for_theme_mode(
        self, theme_name: str, dark: bool
    ) -> ColorSystem:
        """Get color system for theme mode"""
        theme = self.themes[theme_name]

        if dark:
            if "dark" in theme:
                return theme["dark"]
            return theme["light"]

        if "light" in theme:
            return theme["light"]
        return theme["dark"]


theme_manager = ThemeManager()



"""Site Model List View"""
from __future__ import annotations

from textual.widgets import ListView

from parllama.widgets.site_model_list_item import SiteModelListItem


class SiteModelListView(ListView):
    """Site Model List View."""

    def __init__(self, **kwargs) -> None:
        """Initialize the view."""
        super().__init__(**kwargs)

    def on_focus(self, value: bool) -> None:
        """Watch the focus property and select first visible item if none are selected."""
        if value and self.index is None:
            items = [l for l in self.query(SiteModelListItem) if l.display]
            if len(items):
                self.index = self.children.index(items[0])



"""Chat manager class"""
from __future__ import annotations

import datetime
import os
from typing import Any

import simplejson as json
from ollama import Options as OllamaOptions
from textual.app import App

from parllama.messages.main import SessionListChanged
from parllama.messages.main import SessionSelected
from parllama.models.chat import ChatSession
from parllama.models.settings_data import settings


class ChatManager:
    """Chat manager class"""

    app: App[Any]
    sessions: list[ChatSession] = []
    options: OllamaOptions = {}
    current_session: ChatSession | None = None

    def __init__(self) -> None:
        """Initialize the chat manager"""
        self.load_sessions()

    def set_app(self, app: App[Any]) -> None:
        """Set the app"""
        self.app = app

    def mk_session_name(self, base_name: str) -> str:
        """Generate a unique session name"""
        session_name = base_name
        good = self.get_session_by_name(session_name) is None
        i = 0
        while not good:
            if good:
                break
            i += 1
            session_name = f"{base_name} {i}"
            good = self.get_session_by_name(session_name) is None
        return session_name

    def new_session(
        self,
        *,
        session_name: str,
        model_name: str,
        options: OllamaOptions | None = None,
    ) -> ChatSession:
        """Create a new chat session"""
        session = ChatSession(
            session_name=self.mk_session_name(session_name),
            llm_model_name=model_name,
            options=options or self.options,
        )
        self.sessions.append(session)
        self.current_session = session
        self.sort_sessions()
        self.notify_changed()
        return session

    def get_session(self, session_id: str) -> ChatSession | None:
        """Get a chat session"""
        for session in self.sessions:
            if session.session_id == session_id:
                return session
        return None

    def get_session_by_name(self, session_name: str) -> ChatSession | None:
        """Get a chat session by name"""
        for session in self.sessions:
            if session.session_name == session_name:
                return session
        return None

    def delete_session(self, session_id: str) -> None:
        """Delete a chat session"""
        for session in self.sessions:
            if session.session_id == session_id:
                self.sessions.remove(session)
                if self.current_session == session:
                    self.current_session = None
                p = os.path.join(settings.chat_dir, f"{session_id}.json")
                if os.path.exists(p):
                    os.remove(p)
                self.notify_changed()
                return

    def notify_changed(self) -> None:
        """Notify changed"""
        self.app.post_message(SessionListChanged())

    def get_current_session(self) -> ChatSession | None:
        """Get the current chat session"""
        return self.current_session

    def get_or_create_session_name(
        self, *, session_name: str, model_name: str, options
    ) -> ChatSession:
        """Get or create a chat session"""
        session = self.get_session_by_name(session_name)
        if not session:
            session = self.new_session(
                session_name=session_name, model_name=model_name, options=options
            )
        self.current_session = session
        return session

    def set_current_session(self, session_id: str) -> ChatSession | None:
        """Set the current chat session"""
        self.current_session = self.get_session(session_id)
        self.app.post_message(SessionSelected(session_id))
        return self.current_session

    def load_sessions(self) -> None:
        """Load chat sessions from files"""
        for f in os.listdir(settings.chat_dir):
            f = f.lower()
            if not f.endswith(".json"):
                continue
            try:
                with open(
                    os.path.join(settings.chat_dir, f), mode="rt", encoding="utf-8"
                ) as fh:
                    data: dict = json.load(fh)
                    session = ChatSession(
                        session_name=data["session_name"],
                        llm_model_name=data["llm_model_name"],
                        session_id=data["session_id"],
                        messages=data["messages"],
                        options=data.get("options"),
                        last_updated=datetime.datetime.fromisoformat(
                            data["last_updated"]
                        ),
                    )
                    self.sessions.append(session)
            except:  # pylint: disable=bare-except
                self.app.notify(f"Error loading session {f}", severity="error")
        self.sort_sessions()

    def sort_sessions(self) -> None:
        """Sort sessions by last_updated field in descending order."""
        self.sessions.sort(key=lambda x: x.last_updated, reverse=True)


chat_manager = ChatManager()



"""Ollama API Models"""
from __future__ import annotations

from datetime import datetime
from typing import Literal
from typing import TypeAlias

from pydantic import BaseModel

MessageRoles: TypeAlias = Literal["user", "assistant", "system"]


class SiteModel(BaseModel):
    """Ollama Site Model."""

    name: str
    description: str
    url: str
    num_pulls: str
    num_tags: str
    tags: list[str]
    updated: str


class SiteModelData(BaseModel):
    """Ollama Site Model Data."""

    models: list[SiteModel]
    last_update: datetime = datetime.now()


class ModelDetails(BaseModel):
    """Ollama Model Details."""

    parent_model: str
    format: str
    family: str
    families: list[str]
    parameter_size: str
    quantization_level: str


class ModelShowPayload(BaseModel):
    """Ollama Model Show Payload."""

    license: str | None = None
    modelfile: str
    parameters: str | None = None
    template: str
    # details: ModelDetails # omit of being combined with Model


class Model(BaseModel):
    """Ollama Model"""

    name: str
    model: str
    modified_at: datetime
    size: int
    digest: str
    details: ModelDetails
    expires_at: datetime | None = None


class ModelListPayload(BaseModel):
    """List models response."""

    models: list[Model]


class FullModel(Model):
    """Ollama Full Model"""

    license: str | None = None
    modelfile: str
    parameters: str | None = None
    template: str | None = None



"""Session list item"""
from __future__ import annotations

from rich.text import Text
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.widgets import Label
from textual.widgets import ListItem

from parllama.models.chat import ChatSession


class SessionListItem(ListItem, can_focus=False, can_focus_children=True):
    """Session list item"""

    DEFAULT_CSS = """
    SessionListItem {
      height: 4;
      width: 1fr;
      padding-left: 1;
      padding-right: 1;
    }
    """
    session: ChatSession

    def __init__(self, session: ChatSession, **kwargs) -> None:
        """Initialise the view."""
        super().__init__(**kwargs)
        self.session = session

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""
        with Vertical():
            yield Label(self.session.session_name)
            temp = (
                f"{self.session.options['temperature']:.2f}"
                if self.session.options.get("temperature") is not None
                else "Default"
            )

            yield Label(
                Text.assemble(self.session.llm_model_name, " ", f"Temp: {temp}")
            )
            yield Label(
                Text.assemble(
                    "Last updated: ",
                    self.session.last_updated.strftime("%Y-%m-%d %H:%M:%S"),
                )
            )



"""Grid list of local models."""
from __future__ import annotations

import webbrowser

from textual.binding import Binding
from textual.containers import Grid
from textual.reactive import Reactive
from textual.widget import Widget

from parllama.messages.main import LocalModelDeleteRequested
from parllama.messages.main import ShowLocalModel
from parllama.widgets.local_model_list_item import LocalModelListItem


class LocalModelGridList(Grid, can_focus=False):
    """Grid list of local models."""

    BINDINGS = [
        Binding(
            key="right",
            action="select_right",
            show=False,
        ),
        Binding(
            key="left",
            action="select_left",
            show=False,
        ),
        Binding(
            key="up",
            action="select_up",
            show=False,
        ),
        Binding(
            key="down",
            action="select_down",
            show=False,
        ),
        Binding(
            key="ctrl+b",
            action="browser",
            description="Browser",
            show=False,
        ),
        Binding(
            key="enter",
            action="open",
            description="Details",
            show=False,
        ),
        Binding(
            key="delete",
            action="delete_item",
            description="Delete",
            show=False,
        ),
    ]
    DEFAULT_CSS = """
    LocalModelGridList{
        width: 1fr;
        height: auto;
        grid-size: 2;
        grid-columns: 1fr;
        grid-rows: 9;
        align: left top;
    }
    """
    selected: Reactive[LocalModelListItem | None] = Reactive(None)
    old_selected: LocalModelListItem | None = None

    def __init__(self, **kwargs) -> None:
        """Initialise the widget."""
        super().__init__(**kwargs)

    def watch_selected(self, selected: LocalModelListItem | None) -> None:
        """Watch the selected item."""
        if self.old_selected != selected:
            if self.old_selected:
                self.old_selected.set_class(False, "--highlight")
            if selected:
                selected.set_class(True, "--highlight")
                selected.focus()
            self.old_selected = selected

    def select_first_item(self):
        """Select the first item."""
        items = self.query(LocalModelListItem)
        if len(items):
            self.selected = items[0]
        else:
            self.selected = None

    def action_select_left(self) -> None:
        """Select the item to the left."""
        focused = self.selected

        if not focused:
            self.select_first_item()
            return
        items = self.query(LocalModelListItem)
        for current_index, item in enumerate(items):
            if item == focused:
                num_items = len(items)
                if (num_items % 2 == 1) and current_index == num_items - 1:
                    return
                num_cols = 2  # Assuming a 2-column grid
                num_rows = (
                    num_items + num_cols - 1
                ) // num_cols  # Calculate the number of rows
                row = current_index // num_cols  # Calculate the current row
                col = current_index % num_cols  # Calculate the current column

                # Move left one column, wrapping around if necessary
                new_col = (col - 1) % num_cols
                new_index = row * num_cols + new_col

                # Handle the case when there are an odd number of items and the new index is out of bounds
                if new_index < 0:
                    if row == num_rows - 1 and num_items % num_cols == 1:
                        # Last row with an odd number of items, don't change the index
                        new_index = current_index
                    else:
                        new_index = (
                            num_items - 1
                        )  # Move to the last item in the second column

                self.selected = items[new_index]
                return

    def action_select_right(self) -> None:
        """Select the item to the right."""
        focused = self.selected
        if not focused:
            self.select_first_item()
            return
        items = self.query(LocalModelListItem)
        for current_index, item in enumerate(items):
            if item == focused:
                num_items = len(items)
                if (num_items % 2 == 1) and current_index == num_items - 1:
                    return
                num_cols = 2  # Assuming a 2-column grid
                num_rows = (
                    num_items + num_cols - 1
                ) // num_cols  # Calculate the number of rows
                row = current_index // num_cols  # Calculate the current row
                col = current_index % num_cols  # Calculate the current column

                # Move left one column, wrapping around if necessary
                new_col = (col - 1) % num_cols
                new_index = row * num_cols + new_col

                # Handle the case when there are an odd number of items and the new index is out of bounds
                if new_index < 0:
                    if row == num_rows - 1 and num_items % num_cols == 1:
                        # Last row with an odd number of items, don't change the index
                        new_index = current_index
                    else:
                        new_index = (
                            num_items - 1
                        )  # Move to the last item in the second column

                self.selected = items[new_index]
                return

    def action_select_up(self) -> None:
        """Select the item to the top."""
        focused = self.selected
        if not focused:
            self.select_first_item()
            return
        items = self.query(LocalModelListItem)
        for current_index, item in enumerate(items):
            if item == focused:

                num_items = len(items)
                num_cols = 2  # Assuming a 2-column grid
                num_rows = (
                    num_items + num_cols - 1
                ) // num_cols  # Calculate the number of rows

                row = current_index // num_cols  # Calculate the current row
                col = current_index % num_cols  # Calculate the current column

                # Move up one row, wrapping around if necessary
                new_row = (row - 1) % num_rows
                new_index = new_row * num_cols + col

                # Handle the case when there are an odd number of items and the last row has only one item
                if new_index >= num_items:
                    if col == 0:
                        new_index = (
                            num_items - 1
                        )  # Move to the last item in the first column
                    else:
                        new_index = (
                            num_items - 2
                        )  # Move to the last item in the second column

                self.selected = items[new_index]
                return

    def action_select_down(self) -> None:
        """Select the item to the bottom."""
        focused = self.selected
        if not focused:
            self.select_first_item()
            return
        items = self.query(LocalModelListItem)
        for current_index, item in enumerate(items):
            if item == focused:
                num_items = len(items)
                num_cols = 2  # Assuming a 2-column grid
                num_rows = (
                    num_items + num_cols - 1
                ) // num_cols  # Calculate the number of rows

                row = current_index // num_cols  # Calculate the current row
                col = current_index % num_cols  # Calculate the current column

                # Move down one row, wrapping around if necessary
                new_row = (row + 1) % num_rows
                new_index = new_row * num_cols + col

                # Handle the case when there are an odd number of items and the last row has only one item
                if new_index >= num_items:
                    new_index = col  # Move to the first item in the current column

                self.selected = items[new_index]
                return

    def action_delete_item(self) -> None:
        """Delete the selected item."""
        if not self.selected:
            return
        self.post_message(
            LocalModelDeleteRequested(widget=self, model_name=self.selected.model.name)
        )

    def remove_item(self, model_name: str) -> None:
        """Remove a model from the list."""
        for item in self.query(LocalModelListItem):
            if item.model.name == model_name:
                self.action_select_left()
                item.remove()
                return

    def set_item_loading(self, model_name: str, loading: bool) -> None:
        """Set item loading state."""
        for item in self.query(LocalModelListItem):
            if item.model.name in [model_name, f"{model_name}:latest"]:
                item.loading = loading
                return

    def select_by_name(self, model_name: str) -> None:
        """Select item by model name."""
        for item in self.query(LocalModelListItem):
            if item.display and item.model.name == model_name:
                item.focus()
                return

    def action_browser(self) -> None:
        """Open the model in the browser."""
        if not self.selected:
            return
        webbrowser.open(f"https://ollama.com/library/{self.selected.model.name}")

    def action_open(self) -> None:
        """Open the model."""
        if not self.selected:
            return
        self.post_message(ShowLocalModel(self.selected.model))

    def filter(self, value: str) -> None:
        """Filter the list use value from search / filter box."""
        first_visible: Widget | None = None
        num_visible: int = 0
        for item in self.query(LocalModelListItem):
            if not value:
                item.display = True
                item.disabled = False
                continue
            item.display = value.lower() in item.model.name.lower()
            item.disabled = not item.display
            if item.display and not first_visible:
                num_visible += 1
                first_visible = item



"""Main screen for TUI."""
from __future__ import annotations

from typing import cast

from rich.console import RenderableType
from textual import on
from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Footer
from textual.widgets import Header
from textual.widgets import Static
from textual.widgets import TabbedContent
from textual.widgets import TabPane

from parllama.messages.main import ModelInteractRequested
from parllama.messages.main import PsMessage
from parllama.messages.main import StatusMessage
from parllama.models.settings_data import ScreenType
from parllama.models.settings_data import settings
from parllama.widgets.views.chat_view import ChatView
from parllama.widgets.views.create_model_view import ModelCreateView
from parllama.widgets.views.local_model_view import LocalModelView
from parllama.widgets.views.log_view import LogView
from parllama.widgets.views.model_tools_view import ModelToolsView
from parllama.widgets.views.site_model_view import SiteModelView


class MainScreen(Screen[None]):
    """Main screen."""

    BINDINGS = []

    CSS_PATH = "main_screen.tcss"

    status_bar: Static
    ps_status_bar: Static
    tabbed_content: TabbedContent
    log_view: LogView
    local_view: LocalModelView
    site_view: SiteModelView
    model_tools_view: ModelToolsView
    create_view: ModelCreateView
    chat_view: ChatView

    def __init__(self, **kwargs) -> None:
        """Initialize the Main screen."""
        super().__init__(**kwargs)
        self.status_bar = Static("", id="StatusBar")
        self.ps_status_bar = Static("", id="PsStatusBar")
        self.ps_status_bar.display = False

        self.local_view = LocalModelView(id="local_models")
        self.site_view = SiteModelView(id="site_models")
        self.create_view = ModelCreateView(id="model_create")
        self.model_tools_view = ModelToolsView(id="model_tools")
        self.chat_view = ChatView(id="chat_view")
        self.log_view = LogView()

    async def on_mount(self) -> None:
        """Mount the Main screen."""
        self.set_timer(0.5, self.done_loading)

    def done_loading(self) -> None:
        """Hide loading indicator."""
        self.tabbed_content.loading = False

    def compose(self) -> ComposeResult:
        """Compose the Main screen."""
        yield Header(show_clock=True)
        yield Footer()
        yield self.status_bar
        yield self.ps_status_bar

        with TabbedContent(id="tabbed_content", initial=settings.last_screen) as tc:
            self.tabbed_content = tc
            tc.loading = True

            with TabPane("Local", id="Local"):
                yield self.local_view
            with TabPane("Site", id="Site"):
                yield self.site_view
            with TabPane("Tools", id="Tools"):
                yield self.model_tools_view
            with TabPane("Create", id="Create"):
                yield self.create_view
            with TabPane("Chat", id="Chat"):
                yield self.chat_view
            with TabPane("Logs", id="Logs"):
                yield self.log_view

    @on(TabbedContent.TabActivated)
    def on_tab_activated(self, msg: TabbedContent.TabActivated) -> None:
        """Tab activated event"""
        msg.stop()
        # self.notify(f"tab activated: {msg.tab.label.plain}")
        settings.last_screen = cast(ScreenType, msg.tab.label.plain)
        settings.save_settings_to_file()

        self.log_view.richlog.write(f"Tab activated: {msg.tab.label.plain}")

    @on(StatusMessage)
    def on_status_message(self, msg: StatusMessage) -> None:
        """Status message event"""
        # msg.stop()
        self.update_status(msg.msg)
        if msg.log_it:
            self.log_view.richlog.write(msg.msg)

    def update_status(self, msg: RenderableType):
        """Update the status bar."""
        self.status_bar.update(msg)

    @on(PsMessage)
    def on_ps_message(self, msg: PsMessage) -> None:
        """PS status message event"""
        msg.stop()
        self.update_ps_status(msg.msg)

    def update_ps_status(self, msg: RenderableType):
        """Update the ps status bar."""
        self.ps_status_bar.update(msg)
        self.ps_status_bar.display = bool(msg)

    def change_tab(self, tab: ScreenType) -> None:
        """Change active tab."""
        self.tabbed_content.active = tab

    def action_site_tag_clicked(self, model_tag: str) -> None:
        """Update search box with tag"""
        self.query_one("#site_models", SiteModelView).search_input.value = model_tag

    @on(ModelInteractRequested)
    def on_model_interact_requested(self, msg: ModelInteractRequested) -> None:
        """Model interact requested event"""
        msg.stop()



"""Model for application settings."""
from __future__ import annotations

import functools
import os
import shutil
from argparse import Namespace

import ollama
import simplejson as json
from pydantic import BaseModel

from ..utils import get_args
from ..utils import ScreenType
from ..utils import valid_screens


class Settings(BaseModel):
    """Model for application settings."""

    no_save: bool = False
    data_dir: str = os.path.expanduser("~/.parllama")
    cache_dir: str = ""
    chat_dir: str = ""
    settings_file: str = "settings.json"
    theme_name: str = "par"
    starting_screen: ScreenType = "Local"
    last_screen: ScreenType = "Local"
    last_chat_model: str = ""
    last_chat_temperature: float | None = None
    last_chat_session_name: str = "My Chat"
    theme_mode: str = "dark"
    site_models_namespace: str = ""
    max_log_lines: int = 1000
    ollama_host: str = "http://localhost:11434"
    ollama_ps_poll_interval: int = 3
    auto_name_chat: bool = True

    # pylint: disable=too-many-branches
    def __init__(self) -> None:
        """Initialize BwItemData."""
        super().__init__()
        args: Namespace = get_args()

        if args.no_save:
            self.no_save = True

        self.data_dir = (
            args.data_dir
            or os.environ.get("PARLLAMA_DATA_DIR")
            or os.path.expanduser("~/.parllama")
        )
        self.cache_dir = os.path.join(self.data_dir, "cache")
        self.chat_dir = os.path.join(self.data_dir, "chats")

        os.makedirs(self.cache_dir, exist_ok=True)
        os.makedirs(self.chat_dir, exist_ok=True)

        if not os.path.exists(self.data_dir):
            raise FileNotFoundError(
                f"Par Llama data directory does not exist: {self.data_dir}"
            )

        self.settings_file = os.path.join(self.data_dir, "settings.json")
        if args.restore_defaults:
            if os.path.exists(self.settings_file):
                os.unlink(self.settings_file)
            theme_file = os.path.join(self.data_dir, "themes", "par.json")
            if os.path.exists(theme_file):
                os.unlink(theme_file)
        if args.clear_cache:
            if os.path.exists(self.cache_dir):
                shutil.rmtree(self.cache_dir, ignore_errors=True)
                os.makedirs(self.cache_dir, exist_ok=True)

        if args.purge_chats:
            if os.path.exists(self.chat_dir):
                shutil.rmtree(self.chat_dir, ignore_errors=True)
                os.makedirs(self.chat_dir, exist_ok=True)

        self.load_from_file()
        url = os.environ.get("OLLAMA_URL")
        if args.ollama_url:
            url = args.ollama_url
        if url:
            if not (url.startswith("http://") or url.startswith("https://")):
                raise ValueError("Ollama URL must start with http:// or https://")
            self.ollama_host = url

        if os.environ.get("PARLLAMA_THEME_NAME"):
            self.theme_name = os.environ.get("PARLLAMA_THEME_NAME", self.theme_name)

        if os.environ.get("PARLLAMA_THEME_MODE"):
            self.theme_mode = os.environ.get("PARLLAMA_THEME_MODE", self.theme_mode)

        if args.theme_name:
            self.theme_name = args.theme_name
        if args.theme_mode:
            self.theme_mode = args.theme_mode

        if args.starting_screen:
            self.starting_screen = args.starting_screen.capitalize()
            if self.starting_screen not in valid_screens:
                self.starting_screen = "Local"

        if args.ps_poll:
            self.ollama_ps_poll_interval = args.ps_poll
        self.save_settings_to_file()

    def load_from_file(self) -> None:
        """Load settings from file."""
        try:
            with open(self.settings_file, encoding="utf-8") as f:
                data = json.load(f)
                url = data.get("ollama_host", self.ollama_host)

                if url.startswith("http://") or url.startswith("https://"):
                    self.ollama_host = url
                else:
                    print("ollama_host must start with http:// or https://")
                self.theme_name = data.get("theme_name", self.theme_name)
                self.theme_mode = data.get("theme_mode", self.theme_mode)
                self.site_models_namespace = data.get("site_models_namespace", "")
                self.starting_screen = data.get("starting_screen", "Local")
                if self.starting_screen not in valid_screens:
                    self.starting_screen = "Local"

                self.last_screen = data.get("last_screen", "Local")
                if self.last_screen not in valid_screens:
                    self.last_screen = self.starting_screen
                self.last_chat_model = data.get("last_chat_model", self.last_chat_model)
                self.last_chat_temperature = data.get("last_chat_temperature")
                self.last_chat_session_name = data.get(
                    "last_chat_session_name", self.last_chat_session_name
                )
                self.max_log_lines = max(0, data.get("max_log_lines", 1000))
                self.ollama_ps_poll_interval = data.get(
                    "ollama_ps_poll_interval", self.ollama_ps_poll_interval
                )
                self.auto_name_chat = data.get("auto_name_chat", self.auto_name_chat)
        except FileNotFoundError:
            pass  # If file does not exist, continue with default settings

    def save_settings_to_file(self) -> None:
        """Save settings to file."""
        if self.no_save:
            return
        os.makedirs(self.data_dir, exist_ok=True)
        if not os.path.exists(self.data_dir):
            raise FileNotFoundError(
                f"Par Llama data directory does not exist: {self.data_dir}"
            )

        with open(self.settings_file, "w", encoding="utf-8") as f:
            f.write(self.model_dump_json(indent=4))

    def ensure_cache_folder(self) -> None:
        """Ensure the cache folder exists."""
        if not os.path.exists(self.cache_dir):
            os.mkdir(self.cache_dir)

    @functools.cached_property
    def ollama_client(self) -> ollama.Client:
        """Get the ollama client."""
        return ollama.Client(host=self.ollama_host)


settings = Settings()
